
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;

/**
   Implement this task graph using two threads
   and "message passing". This will not deadlock.

              task1
             /     \
            /       \
       task2         task3
            \       /
             \     /
              task4
*/
public class TaskGraph_v2_v2
{
   public static void main(String[] args)
   {
      final Task<Integer,Integer> task1 = new Task<>(1, 1, 2);
      final Task<Integer,Integer> task2 = new Task<>(2, 2, 1);
      final Task<Integer,Integer> task3 = new Task<>(3, 3, 3);
      final Task<Integer,Integer> task4 = new Task<>(4, 4, 2);

      final CompletableFuture<Integer> future1 = new CompletableFuture<>(); // empty future
      final CompletableFuture<Integer> future2 = new CompletableFuture<>(); // empty future

      final Thread t1 = new Thread(
                          ()->{try{ task1.run();
                                    future1.complete(0); // Send a message.
                                    task2.run();
                                    future2.get();       // Receive a message.
                                    task4.run();
                                  }catch(InterruptedException
                                        |ExecutionException e){} }
                                  );

      final Thread t2 = new Thread(
                          ()->{try{ future1.get();       // Receive a message.
                                    task3.run();
                                    future2.complete(0); // Send a message.
                                  }catch(InterruptedException
                                        |ExecutionException e){} }
                                  );
      t2.start();
      t1.start();
   }
}
