
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;

/**
   Implement this task graph using four threads,
   forks, and "message passing". This will not deadlock.

   Compare this version with TaskGraph_v1.java which uses fork/join.

              task1
             /     \
            /       \
       task2         task3
            \       /
             \     /
              task4
*/
public class TaskGraph_v3
{
   public static void main(String[] args)
   {
      final Task<Integer,Integer> task1 = new Task<>(1, 1, 2);
      final Task<Integer,Integer> task2 = new Task<>(2, 2, 1);
      final Task<Integer,Integer> task3 = new Task<>(3, 3, 3);
      final Task<Integer,Integer> task4 = new Task<>(4, 4, 2);

      final CompletableFuture<Integer> future = new CompletableFuture<>(); // empty future

      final Thread t4 = new Thread(task4);

      final Thread t3 = new Thread(
                          ()->{ task3.run();
                                future.complete(0); } // Send a message to taks2.
                                  );

      final Thread t2 = new Thread(
                          ()->{try{ task2.run();
                                    future.get();  // Receive a message from task3.
                                    t4.start();    // fork
                                  }catch(InterruptedException
                                        |ExecutionException e){} }
                                  );

      final Thread t1 = new Thread(
                          ()->{ task1.run();
                                t2.start();    // fork
                                t3.start(); }  // fork
                                  );
      t1.start();
   }
}
