
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;

/**
   Implement this task graph on a thread pool.
   This implementation uses "message passing"
   and task submissions within the tasks (forks).
   This can deadlock.

   Compare this version with TaskGraph_v3.java.

              task1
             /     \
            /       \
       task2         task3
            \       /
             \     /
              task4
*/
public class TaskGraph_v5
{
   public static void main(String[] args)
   {
      final int threads = 3;
      final MyThreadPool_v1 executor = new MyThreadPool_v1(threads);

      final Task<Integer,Integer> task1 = new Task<>(1, 1, 2);
      final Task<Integer,Integer> task2 = new Task<>(2, 2, 1);
      final Task<Integer,Integer> task3 = new Task<>(3, 3, 3);
      final Task<Integer,Integer> task4 = new Task<>(4, 4, 2);

      final CompletableFuture<Integer> future = new CompletableFuture<>(); // empty future

      final Runnable r3 = ()->{ task3.run();
                                future.complete(0); // Send a message to task2.
                              };

      final Runnable r2 = ()->{try{ task2.run();
                                    future.get();  // Receive a message from taks3.
                                    executor.execute(task4);  // fork
                                    executor.shutdown();
                                  }catch(InterruptedException
                                        |ExecutionException e){}
                              };

      final Runnable r1 = ()->{ task1.run();
                                executor.execute(r2);  // fork
                                executor.execute(r3);  // fork
                              };

      executor.execute( r1 );
   }
}
