
import java.util.concurrent.Future;
import java.util.concurrent.Executors;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.ExecutionException;

/**
   Implement this task graph on an executor service
   style thread pool using Callable tasks and Futures.

              task1
             /     \
            /       \
       task2         task3
            \       /
             \     /
              task4

   It is very hard to cause a deadlock
   when using this style of programming.

   If we break the rule that
      "Futures must always be declared final"
   then we can cause a deadlock.
*/
public class TaskGraph_v6b
{
   // These static declarations make the compiler do something it shouldn't do!
   private static Future<Integer> future2 = null; // assign future2 once
   private static Future<Integer> future3 = null; // assign future3 once
   // Or, try these two declarations.
   //private static Future<Integer> future2 = new java.util.concurrent.CompletableFuture<Integer>();
   //private static Future<Integer> future3 = new java.util.concurrent.CompletableFuture<Integer>();

   public static void main(String[] args)
   {
      final int threads = 1; // try 2 threads
      final var executor = new MyExecutorService<Integer>(threads);
    //final ExecutorService executor = Executors.newFixedThreadPool(threads);

      final TaskC<Integer,Integer> task1 = new TaskC<>(1, 1, 2);
      final TaskC<Integer,Integer> task2 = new TaskC<>(2, 2, 1);
      final TaskC<Integer,Integer> task3 = new TaskC<>(3, 3, 3);
      final TaskC<Integer,Integer> task4 = new TaskC<>(4, 4, 2);

      // The compiler will not accept these declarations because they are not final.
      //Future<Integer> future2 = null;
      //Future<Integer> future3 = null;

      final Future<Integer> future1 = executor.submit(
                          ()->{ final Integer i = task1.call();
                                future2 = executor.submit(task2); // reassign future2
                                future3 = executor.submit(task3); // reassign future3
                                return i; });

      final Future<Integer> future4 = executor.submit( // This task is in the task queue ahead of task2.
                          ()->{ future2.get();
                                future3.get();
                                return task4.call(); });

      try{ future4.get(); }
      catch(InterruptedException|ExecutionException e){}

      executor.shutdown();
   }
}
